import typer
import re
from y4d import codeParser as codeParser

app = typer.Typer()

#This function is responsible for converting classes and functions to comments in a .cpp or .h file
def commentMaker(source: str, type: str, name: str, isolate =0, option=0, output = ""):

    #Find where the classes or functions start
    position = codeParser.positions(source, type , name, option)
    if position == ["error"]:
        return
    if position == []:
        if isolate == 0:
            print("NOT FOUND")
        return
    
    num = len(position)
    
    #Get all the code to edit it
    with open(source, 'r') as f:
        lines = f.readlines()
        
    for i in range(0, int(num), 4):
        lines[position[i]-1] = "//" + lines[position[i]-1]
        for line in range(position[i], position[i+1]):
            lines[line] = "//" + lines[line]
    
    
    #Write the modified code into the file.
    if output != "":
        with open(output, 'w') as f:
            f.writelines(lines)
    else:
        for line in lines:
            print(line)



#This function is responsible for removing all comments from a .cpp or .h file
@app.command("d")
def deleteComments(source: str  = typer.Argument(..., help="The path of the .cpp or .h file the user wants delete comments to work on."),
                    output: str  = typer.Option("", "-o", help="The path of the .cpp or .h file the user wants the output to be saved in (Default is printing on the terminal).")):
    """
    Deletes all the comments from a .cpp or .h file and output the source code into the file of your choosing.
    """
    with open(source, 'r') as f:
        source = f.read()
    
    # Remove all comments from the input file
    source = re.sub(r'\/\/.*?$|\/\*[\s\S]*?\*\/', '', source, flags=re.MULTILINE)

    # Write the modified contents to the output file
    if output != "":
        with open(output, 'w') as f:
            f.write(source)
    else:
        print(source)



#This function is responsible for extracting all comments from a .cpp or .h file into a file or to print on the terminal
@app.command("e")
def extractComments(source: str  = typer.Argument(..., help="The path of the .cpp or .h file the user wants extracts comments to work on."),
                    output: str  = typer.Option("", "-o", help="The path of the .cpp or .h file the user wants the output to be saved in (Default is printing on the terminal).")):
    """
    Extracts all the comments from a C++ file and output the comments into the file of your choosing.
    """
    with open(source, 'r') as f:
        source = f.read()
    
    # Extract all comments from the input file
    comments = re.findall(r'\/\/.*?$|\/\*[\s\S]*?\*\/', source, flags=re.MULTILINE)

    # Write the comments to the output file or to the terminal
    if output != "":
        with open(output, 'w') as f:
            for comment in comments:
                size = len(comment)
                if comment[0:2] == "/*":
                    comment = comment[2:size-2].strip()
                else:
                    comment = comment[2:size].strip()
                f.write(comment + "\n")
    else:
        for comment in comments:
            print(comment)



#This function is responsible for extracting all header comments from a .cpp or .h file into a file or to print on the terminal
@app.command("h")
def extractHeader(input: str  = typer.Argument(..., help="The path of the .cpp or .h file the user wants extracts header comments to work on."),
                   output: str  = typer.Option("", "-o", help="The path of the .cpp or .h file the user wants the output to be saved in (Default is printing on the terminal).")):
    """
    Extracts all header comments from a C++ file and output the comments into the file of your choosing.
    """
    with open(input, 'r') as f:
        source = f.read()

    #Find all header comments
    header = re.findall(r'\/\/.*?$|\/\*[\s\S]*?\*\/', source, flags=re.MULTILINE)
    amount = 0

    while source[0:1] == '\n':
        source = source[1:]

    while source[0:2] == "//" or source[0:2] == "/*":
        source = re.sub(r'\/\/.*?$|\/\*[\s\S]*?\*\/', '',source, flags=re.MULTILINE, count=1)
        amount += 1
        while source[0:1] == '\n':
            source = source[1:]

    # Write the header comments to the output file
    if output != "":
        with open(output, 'w') as f:
            for n in range(amount):
                comment = header[n]
                size = len(comment)
                if comment[0:2] == "/*":
                    comment = comment[2:size-2]
                else:
                    comment = comment[2:size]
                f.write(comment + "\n")
    else:
        for n in range(amount):
            comment = header[n]
            print(comment)



#This function is responsible for commenting out classes specified by the user from a .cpp or .h file
@app.command("c")
def CommentOutClass(
    source: str  = typer.Argument(..., help="The path of the .cpp or .h file the user wants comment out class to work on."),
    prototype: str  = typer.Argument(..., help="The class the user wants to check (Must input like this: \"class name\")."),
    all: str  = typer.Option("f", "-all", help="If t this will comment out the class with all its children classes, if c this will comment out the class itself without any member function implemented outside the class,if f this will comment out the class with its member functions implemented outside the class. Takes c, t, or f. Default value is f."),
    isolate: int = typer.Argument(0, hidden=True, help="A hidden variable for developers' use, used to show that function was called by isolator."),
    output: str  = typer.Option("", "-o", help="The path of the .cpp or .h file the user wants the output to be saved in (Default is printing on the terminal).")):
    """
    Comments out a class implementation from a C++ file using a class prototype (equivalent to deleting the class).
    """
    if all.lower() not in ["f","t", "c"]:
        print("Invalid input for -all")
        return False
    
    type = prototype.split(" ")[0]
    option = 1
    if all == "t":
        option = 0
    elif all == "c":
        option = -1

    commentMaker(source, type , prototype, isolate, option, output)



#This function is responsible for commenting out functions specified by the user from a .cpp or .h file
@app.command("f")
def CommentOutFunction(
    source: str  = typer.Argument(..., help="The path of the .cpp or .h file the user wants comment out class to work on."),
    prototype: str  = typer.Argument(..., help="The function the user wants to check (Must input like this: \"int functionName(int, int)\"."),
    isolate: int = typer.Argument(0, hidden=True, help="A hidden variable for developers' use, used to show that function was called by isolator."),
    output: str  = typer.Option("", "-o", help="The path of the .cpp or .h file the user wants the output to be saved in (Default is printing on the terminal).")):
    """
    Comments out a function implementation from a C++ file using a function prototype (equivalent to deleting the function).
    """
    commentMaker(source, "function" , prototype, isolate, 0, output)



#This function is responsible for removing all comments from a .cpp or .h file
def deleteForDeveloper(source: str):
    """
    Deletes all the comments from a .cpp or .h file and return the source code.
    """
    # Remove all comments from the input file
    source = re.sub(r'\/\/.*?$|\/\*[\s\S]*?\*\/', '', source, flags=re.MULTILINE)

    return source



@app.callback()
def main():
    """
    Allows the user to control comments by extracting or removing comments, or by commenting out certain classes or functions.
    For more information about each tool simply add the command you need followed by --help, for example:\n\n commentCtrl f --help
    """



if __name__ == "__main__":
    app()



import json
import clang.cindex
from clang.cindex import CursorKind
import re
from pathlib import Path

def findLocationFunction(data, prototype: str, source):
    #Check for static, virtual, pure virtual, and constructors
    is_static = False
    is_pure = False
    is_virtual = False
    #Check static functions
    if len(prototype.split("static")) > 1:
        is_static= True
        prototype = prototype.split("static")[1]

    #Check virtual and pure virtual functions
    if len(prototype.split("virtual")) > 1:
        is_virtual= True
        if len( prototype.split("=0")) > 1:
            prototype = prototype.split("=0")[0]
            is_pure= True
        elif len( prototype.split("= 0")) > 1:
            prototype = prototype.split("= 0")[0]
            is_pure= True
        prototype= prototype.split("virtual")[1]
            
    #Label function by its type 
    if len(prototype.split("template"))> 1 :
        if len(prototype.split("::")) > 1 :
            type = "template_member_function"
        else:
            type = "template_function"
    elif len(prototype.split("::")) > 1:
        type = "member_function"
    else:
        type="function"
    #Extract information from function prototype
    if type == "member_function" or type == "template_member_function":
        parent_class = prototype.split("::")[0].strip().split(" ")[-1]
        returntype = prototype.split(parent_class)[0]
        if type == "template_member_function":
            returntype = returntype.split(">")[1]
        prototype=prototype.replace(" ", "")
        name = prototype.split("::")[1].split("(")[0]
        if len(name.split("~")) > 1 and name.split("~")[1] == parent_class:
            type = "decstructor"
        if name == parent_class:
            type = "constructor"   
    else:
        if type == "template_function":
            prototype = prototype.split(">")[1].strip()
        name = prototype.split('(')[0].split(" ")[-1].strip()
        returntype = prototype.split(name)[0].strip()
        
    if type == "constructor" or type == "decstructor":
        returntype = "void"
        params = "(" + prototype.split("::")[1].split("(")[1]
    else:
        params = prototype.split(name)[1]
        
    qualtype = returntype + params
    qualtype = qualtype.replace(" ", "")

    #Retrieve position of function
    pos=[]
    if type == "function":
        for item in data['nodes']:
            if item['kind'] == "FUNCTION_DECL" and item['spelling'].replace(" ", "") == name and item['prototype'].replace(" ", "") == qualtype:
                if is_static == item['is_static_method'] and is_virtual == item['is_virtual_method'] and is_pure == item['is_pure']:
                    end = item['end']
                    start = item['start']
                    if name == "main":
                        pos += [start, end, "main", item['access_type']]
                    else:
                        pos += [start, end, type, item['access_type']]
                    
    if type == "member_function":
        for item in data['nodes']:
            if item['kind'] == "CXX_METHOD" and item['spelling'].replace(" ", "") == name and item['prototype'].replace(" ", "") == qualtype and item['parent_class']!="" and parent_class == item['parent_class'].split(" ")[1]:
                if is_static == item['is_static_method'] and is_virtual == item['is_virtual_method'] and is_pure == item['is_pure']:
                    end = item['end']
                    start = item['start']
                    pos += [start, end, type, item['access_type']]          
    if type == "template_function" or type == "template_member_function":
        for item in data['nodes']:
            if (item['kind'] == "FUNCTION_TEMPLATE" and item['spelling'].replace(" ", "") == name and item['prototype'].replace(" ", "") == qualtype and is_static == item['is_static_method'] and ((type == "template_function" and item['access_type'] == "invalid") or ( type == "template_member_function" and item['parent_class'] != "" and parent_class == item['parent_class'].split(" ")[1])  )  ):
                    start = item['start']
                    end =-1
                    with open(source, "r") as source_file:
                        lines = source_file.readlines()    
                    #Check for forward declaration 
                    openb =0
                    for char in lines[start-1]:
                        if char == '{':
                            openb = openb+1
                    if openb == 0:
                        pos += [start, start, type, item['access_type']]
                        continue
                    #Find where the function ends 
                    i = start 
                    while(True):
                        for char in lines[i]:
                            if char == '{':
                                openb = openb+1
                            if char == '}':
                                openb = openb -1
                                if openb == 0:
                                    end = i+1
                                    break
                        i = i+1
                        if end != -1:
                            break
                    pos += [start, end, type, item['access_type']] 
    if type == "constructor":
        for item in data['nodes']:
            if item['kind'] == "CONSTRUCTOR" and item['spelling'].replace(" ", "") == name and item['prototype'].replace(" ", "").split(")")[0]+")" == qualtype:
                end = item['end']
                start = item['start']
                pos += [start, end, type, item['access_type']] 
    if type == "decstructor":
        for item in data['nodes']:
            if item['kind'] == "DESTRUCTOR" and item['spelling'].replace(" ", "") == name:
                end = item['end']
                start = item['start']
                pos += [start, end, type, item['access_type']] 
        
    return pos

def findLocationClass(data, prototype: str, source, type: str, iteration = 0):
    #classes list stores all classes that inherits from\friend with main class in order to parse their member functions and nested classes  
    classes =  [prototype]
    pos=[]
    for class_i in classes:
        name = class_i.split(" ")[1]
        class_start = -1
        class_end = -1
        flag=0
        i=0 
        while i < len(data['nodes']):
            item = data['nodes'][i]
            #find start and end positions of a class given its name
            if item['spelling'] == name:
                if ((item['is_class'] == True and type == "class") or (item['is_struct'] == True and type == "struct")):
                    start = item['start']
                    end = item['end']
                    class_start = start 
                    class_end = end
                    pos+=[start, end, type, ""]
                    flag =1
                elif item['kind'] == "CLASS_TEMPLATE":
                    start = item['start']
                    # in-case the template definition was on the previous line
                    if item['col'] <= 12:
                        start -= 1
                    end = item['end']
                    class_start = start 
                    class_end = end
                    pos+=[start, end, type, ""]
                    flag =1
            if flag  == 0 : 
                i = i+1
                continue
            elif iteration == -1:
                return pos  
            #check member functions implemented outside the class
            if item['start'] < class_start or item['end'] > class_end:
                if item['mangled_name'].startswith('?'):
                    mangledName= item['mangled_name'].split("@")
                    if len(mangledName) > 1 and mangledName[1] == name:
                        if item['kind'] == "CXX_METHOD":
                            returnType = item['prototype'].split("(")[0].strip()
                            func_prototype = returnType +" " + name + "::" +item['displayname']
                            pos+= findLocationFunction(data, func_prototype, source)
                    if item['kind'] == "CONSTRUCTOR" or item['kind'] =="DESTRUCTOR":
                        returnType = "void"
                        func_prototype = name+"::" +item['displayname']
                        pos+= findLocationFunction(data, func_prototype, source)
            #check template member functions  
            if item["kind"] == "FUNCTION_TEMPLATE":
                start_line = item['start']
                j = i+1
                template_node = data['nodes'][j]
                while  j < len(data['nodes']) and template_node['start'] == start_line:
                    if template_node['kind'] == "TEMPLATE_TYPE_PARAMETER":
                        template = template_node['spelling']
                    if template_node['kind'] == "TYPE_REF" and template_node['spelling'] == prototype:
                        returnType = item['prototype'].split(" ")[0]
                        func_prototype = "template <typename " + template + "> " + returnType +" " + name + " :: " +item['displayname']
                        pos += findLocationFunction(data, func_prototype, source)
                    template_node = data['nodes'][j]    
                    j+=1 
            #find start and end positions of all classes that inherits from this class       
            for inherit in item['inherits_from']:
                if inherit == name:
                    start = item['start']
                    end = item['end']
                    classes += [type + " "+  item['spelling']]            
            #find start and end positions of all classes with friendship with this class
            for friend in item['friend_with']:
                if friend == class_i:
                    start = item['start']
                    end = item['end']
                    classes +=  [type + " "+  item['spelling']]
            i = i+1
        if iteration == 1:
            return pos

    return pos

#Compile code and return parse tree
def prepareData (source: str):
    
    #Get Code from source
    with open(source, 'r') as file:
        source = file.read()

    #String manipulate source for parse to work (removing libraries and using namespace)
    source = re.sub(r'^#include\s*[\s\S][<"]\S+[>"]$', r'// \g<0>', source, flags=re.MULTILINE)
    source = re.sub(r'^\s*using\s+namespace\s+\S+;$', r'// \g<0>', source, flags=re.MULTILINE)

    index = clang.cindex.Index.create()
    tu = index.parse('dud.cpp', unsaved_files=[('dud.cpp', source)])
        
    output = {"nodes": []}
    friendFlag = False
    classPointer = -1
    for node in tu.cursor.walk_preorder():
        access_type =""
        parent_class = ""
                
        #Save access type of member functions, anything else will have value "invalid"              
        access_type = str(node.access_specifier).split(".")[1]
        #Save name of parent class
        if access_type == "INVALID":
            parent_class = ""
        else:
            parent = node.semantic_parent
            if parent is None:
                parent_class=""
            elif parent.kind == clang.cindex.CursorKind.CLASS_DECL:
                parent_class = "class " + str(parent.spelling)
            elif parent.kind == clang.cindex.CursorKind.STRUCT_DECL:
                parent_class = "struct " + str(parent.spelling)
            
        if node.kind == clang.cindex.CursorKind.CLASS_DECL or node.kind == clang.cindex.CursorKind.STRUCT_DECL:
            classPointer = 0
            friendFlag = False
            
        elif node.kind == clang.cindex.CursorKind.CXX_BASE_SPECIFIER:
            inh = node.spelling
            output["nodes"][classPointer]["inherits_from"].append(inh)

        elif node.kind == clang.cindex.CursorKind.FRIEND_DECL:
            friendFlag = True

        elif friendFlag:
            friend = node.spelling
            if len(friend.split("class")) == 1 or len(friend.split("struct")) ==1:
                friend = node.type.spelling.split('(')[0] + node.displayname
            output["nodes"][classPointer]["friend_with"].append(friend)
            friendFlag = False

        classPointer = classPointer - 1

        #Backbone of the project, retuning parse tree information.
        node_dict = {
            "kind": str(node.kind.name),
            "spelling": node.spelling,
            "prototype": node.type.spelling,
            "displayname": node.displayname,
            "type": str(node.type),
            "start": int(node.location.line),
            "end": int(node.extent.end.line),
            "col": int(node.location.column),
            "mangled_name": node.mangled_name,
            "is_const_method": node.is_const_method(),
            "is_static_method": node.is_static_method(),
            "is_virtual_method": node.is_virtual_method(),
            "is_pure": node.is_pure_virtual_method(),
            "is_struct": node.kind == clang.cindex.CursorKind.STRUCT_DECL,
            "is_class": node.kind == clang.cindex.CursorKind.CLASS_DECL,
            "inherits_from": [],
            "friend_with": [],
            "access_type" : access_type.lower(),
            "parent_class" : parent_class
        }
        output["nodes"].append(node_dict)
    
    jsonFormat = json.dumps(output, indent=4)
    data = json.loads(jsonFormat)
    return data


#Return postions of anything the user is searching for.
def positions ( source: str, type: str, prototype: str, option = 0):
    source_path = Path(source)
    if source_path.exists() == False:
        print("Error: " + source + " doesn't exist")
        return ["error"]
    
    #Checking if it is a .h file and converting it to .cpp, due to libclang requirements
    file_extension = source_path.suffix
    if file_extension == ".h":
        file_name = source_path.stem
        with open(source, 'r') as source:
            content = source.read()
        destination_file = file_name + ".cpp"
        with open(destination_file, 'w') as destination:
            destination.write(content)
        source = destination_file
        
    #Compiling the code
    data = prepareData(source)
    
    #Varibale for saving position of anything the user is searching for.
    pos =[]
    
    if type == "class" or type == "struct":
        pos = findLocationClass(data, prototype , source, type, option)  
        
    if type == "function":
        pos = findLocationFunction( data, prototype, source)  
        
    return pos


import typer
from y4d import codeParser as codeParser
from y4d import commentController as commentController

app = typer.Typer()



@app.command("f")
def isolateFunction(source: str  = typer.Argument(..., help="The source code that the function will be isolated from (.cpp/.h)"),
                    destination: str  = typer.Argument(..., help="The destination code where the function will be embedded or replaced (.cpp/.h)"),
                    prototype: str  = typer.Argument(..., help="The function prototype: return-type function-name parameters-types. In the case of member functions return-type class-name::function-name parameters-types. Example: int test(int, string). Must put them in quotations when using CLI.")):
    """
    Isolates out a function, the function will be taken from source and replace the one in destination or add one.
    """
    
    findFunction = codeParser.positions(source,"function", prototype)
    if findFunction == ["error"]:
        return
    if findFunction == []:
        print("Warning: Function doesn't exist in " + source +" file")
        return 
    
    # Open the source file and read its contents
    with open(source, "r") as source_file:
        lines = source_file.readlines() 
    
    #check if its a member function and parent class exists in destination file 
    if findFunction[2] != "function" and findFunction[2] != "template_function" and findFunction[2] != "main":
        parent_class = prototype.split("::")[0].strip().split(" ")[-1]
        findClass = codeParser.positions(destination, "class", "class " + parent_class)
        if findClass == ["error"]:
            return
        if findClass == []:
            findClass = codeParser.positions(destination, "struct", "struct " + parent_class)
            if findClass == ["error"]:
                return
            if findClass == []:
                print("Warning: Parent Class doesn't exist in " +destination +" file")
                return 
        class_end = findClass[1]
        #fixing protoype
        if findFunction[2] == "constructor" or findFunction[2] == "destructor":
            forward_implementation = prototype.split("::")[1]
        else:
            forward_implementation = prototype.split(parent_class)[0] + prototype.split(parent_class)[1]
            forward_implementation = forward_implementation.split("::")[0] + forward_implementation.split("::")[1]
    
    #commenting out the function in destination file
    commentController.CommentOutFunction(destination, prototype, 1, destination)
    
    #case: forward declaration or memebr functions implemented outside a class
    i=0
    if len(findFunction) > 4:
        i=4
    start_line = findFunction[i+0]
    end_line = findFunction[i+1]
    
    # Extract the lines you want to copy
    implementation = lines[start_line - 1:end_line]
    
    # Open the destination file and insert the copied lines at the appropriate position
    with open(destination, "r") as destination_file:
        lines = destination_file.readlines()
    #functions insertion
    if findFunction[2] == "function" or findFunction[2] == "template_function":
        lines = [prototype +";\n"]+ lines[0:] + ['\n'] + implementation  
    elif findFunction[2] == "main":
        lines = lines[0:] + ['\n'] + implementation  
    #member functions insertion 
    else:
        #if function implemented outside the class 
        if  i == 4:
            lines = lines[0: class_end-1] + [findFunction[i+3] + ": "] + [forward_implementation + ";\n"] + lines[class_end-1: ] +['\n']+ implementation 
        #if function implemented inside the class
        else:
            lines = lines[0: class_end-1] + [findFunction[i+3] + ": "] + implementation + ["\n"] + lines[class_end-1: ]
       
    # Write the modified lines to the destination file
    with open(destination, "w") as destination_file:
        destination_file.writelines(lines)



@app.command("c")
def isolateClass(
    source: str  = typer.Argument(..., help="The source code that the function will be isolated from (.cpp/.h)"), 
    destination: str  = typer.Argument(..., help="The destination code where the function will be embedded or replaced (.cpp/.h)"), 
    prototype: str  = typer.Argument(..., help="The class prototype: class class-name. Example: class test. Must put them in quotations when using CLI."),
    all: str  = typer.Option("f", "-all", help="If t this will isolate the class with all its children classes, if c this will isolate the class itself without any member function implemented outside the class,if f this will isolate the class with its member functions implemented outside the class. Takes c, t, or f. Default value is f.")):
    """
    Isolates out a class, the class will be taken from source and replace the one in destination or add one.
    """
    if all.lower() not in ["f","t", "c"]:
        print("Invalid input for -all")
        return False
    
    option =1
    if all == "t":
        option = 0
    elif all == "c":
        option = -1
    type = prototype.split(" ")[0]
    position = codeParser.positions(source, type, prototype, option)
    
    if position == ["error"]:
        return
    
    #checking if class exists in source code
    if position == []:
        print("Class not found in " + source)
        return
    
    with open(source, "r") as source_file:
        lines = source_file.readlines()
        
    n = len(position)
    things = []
    for i in range(0, int(n), 4): 
        if position[i+2] == type:
            class_start = position[i]
            class_end = position[i+1]
        start_line = position[i]
        end_line = position[i+1]
        #functions implemented outside a class
        if start_line <= class_start or end_line >= class_end:
            things += lines[start_line - 1:end_line]
    
    #checking if class already exists in destination code and commenting it out
    class_position = codeParser.positions(destination, type, prototype, 1)
    if class_position == ["error"]:
        return
    commentController.CommentOutClass(destination, prototype, all, 1, destination)
    
    with open(destination, "r") as destination_file:
        lines = destination_file.readlines()
        
        #if class doesnt exist in destination.cpp, the isolated class will be added at the end of the code.
        if class_position == []:
            lines =  lines +  ["\n"]  + things
        #if class exists in destination.cpp, it will be replaced by the isolated class.
        else:
            lines =  lines[0: class_position[0]-1] + things +  ["\n"] + lines[class_position[0]-1: ]
    
    
    
    # Write the modified lines to the destination file
    with open(destination, "w") as destination_file:
        destination_file.writelines(lines)



@app.callback()
def main():
    """
    Allows the user to copy a given function or class from the source code file into the destination code file and replaces. It consists of two parts, IsolateFunction and IsolateClass.\n\n
    For more information about each tool simply add the command you need followed by --help, for example:\n\n isolate f --help
    """



if __name__ == "__main__":
    app()



import yaml
import json
import typer
import re
from y4d import codeParser as codeParser
from y4d import commentController as commentController

app = typer.Typer()

#Returns everything in the scope the user defined
def scopeGetter(source:str, scope:str ):
    #Checks if scope for function or class
    type = scope.split(" ")[0]
    if type != "class":
        type = "function"
    
    #Find where function or class is
    pos = codeParser.positions(source, type, scope)
    if pos == ["error"]:
        return ["error"]
    if pos == []:
        print("Warning: " + type + " doesn't exist in " + source +" file")
        return ["error"]
    searchPos = []
    count = 0

    for p in pos:
        if count != 2 and count != 3:
            searchPos.append(p)
            count += 1
        elif count == 2:
            count += 1
        else:
            count = 0

    #Get back only the lines in the specified scope
    with open(source, 'r') as f:
        source = f.readlines()
    scopeLines = []
    i = 0
    while i < len(searchPos):
        j = searchPos[i] - 1
        while j < searchPos[i+1]:
            scopeLines.append(source[j])
            j += 1
        i += 2

    source = '\n'.join(scopeLines)
    return source



#Function that uses all other restriction functions that are below it, it recieves a YAML file which lets it know which functions to run
@app.command("r")
def restrict(source: str  = typer.Argument(..., help="The path of the .cpp or .h file the user wants restrictor to work on."),
             rules: str  = typer.Argument(..., help="The path of the YAML file containing user requirements."),
             output: str  = typer.Option("n", "-o", help="If n this will make restrict print the number of violations, Input V if you want a list of violations to be printed and more information (default is n) (Takes only V or n or N)."),
             hide: bool = typer.Argument(False, hidden=True, help="A hidden variable for developers use, used with checkAPI to show the names of the functions or classes that are violating the YAML file (extra)."),
             hide2: bool = typer.Argument(False, hidden=True, help="A hidden variable for developers use, used to show that checkAPI called restrict")):
    """
    Recieves a YAML file made by the user and restrict a .cpp or .h file according to that YAML file, it will return a list of findings (for YAML file explanation check GitHub).
    """
    #Used to control the style of output
    if output not in ["n", "N", "V", "v"]:
        print("Invalid -o input")
        return False
    
    outputBool = False
    if output == "n" or output == "N":
        outputBool = True

    with open(rules) as file:
        yamlFile = yaml.load(file, Loader=yaml.FullLoader)
    jsonFormat = json.dumps(yamlFile, indent=4)
    jsonData = json.loads(jsonFormat)

    #Used to remove comments from file, important for all restrictors (library only one that needs to work without it)
    newSource = "commentDeletedFileForRestrictor.cpp"
    commentController.deleteComments(source, newSource)

    #Following variables are needed to print everything in a readable manner for the user and for the loop to function correctly
    critOld = "start"
    critAns = True
    exactCount = 0
    compareCount = 0
    parentCheck = []
    violationCount = 0

    for criteria, critData in jsonData.items():
        #The following code is to print the findings of the code for each criteria in the YAML file
        if critOld != "start":
            if exactCount != 0 and compareCount != exactCount:
                critAns = critAns & False
            # if critData['restriction'].lower() == 'exactly' and not critAns and not outputBool and not hide:
            #     print(critOld + " are not exactly the same.")
            critAns = True
            compareCount = 0
            exactCount = 0

        critOld = criteria

        #Handles libraries in YAML file
        if criteria == 'libraries':
            for lib in critData['names']:
                if critData['restriction'].lower() != 'exactly':
                    if not libRestrict(source, critData['restriction'], lib, True):
                        critAns = False
                        violationCount += 1
                        if not outputBool:
                            if critData['restriction'].lower() == "forbidden":
                                print("Violating library: " + lib)
                            elif not hide:
                                print("Missing library: " + lib)
                            else:
                                print("Extra library: " + lib)
                else:
                    if not libRestrict(source, 'at_least', lib, True):
                        critAns = False
                        violationCount += 1
                        if not outputBool:
                            if not hide:
                                print("Missing library: " + lib)
                            else:
                                print("Extra library: " + lib)
                    exactCount += 1
                    if exactCount == 1:
                        with open(source, 'r') as f:
                            data = f.read()
                        compareCount =  len(re.findall(r'^#include\s*[\s\S][<"]\S+[>"]$', data, flags=re.MULTILINE))

        #Handles keywords in YAML file
        elif criteria == 'keywords':
            keyCount = True
            for kword in critData['names']:
                if critData['restriction'].lower() != 'exactly':
                    if not wordRestrict(newSource, critData['restriction'], kword, critData['scope'], True):
                        critAns = False
                        violationCount += 1
                        if not outputBool:
                            if critData['restriction'].lower() == "forbidden":
                                print("Violating keyword: " + kword)
                            elif not hide:
                                print("Missing keyword: " + kword)
                            else:
                                print("Extra keyword: " + kword)
                else:
                    if keyCount:
                        print("exactly not supported for keywords, at_least will be used instead.")
                        keyCount = False
                    if not wordRestrict(newSource, 'at_least', kword, critData['scope'], True):
                        critAns = False
                        violationCount += 1
                        if not outputBool:
                            if not hide:
                                print("Missing keyword: " + kword)
                            else:
                                print("Extra keyword: " + kword)

        #Handles classes in YAML file
        elif criteria == 'classes':
            for cls in critData['names']:
                if critData['restriction'].lower() != 'exactly':
                    if not classRestrict(newSource, critData['restriction'], cls, critData['scope'], True):
                        if len(cls.split("::")) != 1:
                            continue
                        critAns = False
                        violationCount += 1
                        if not outputBool:
                            if critData['restriction'].lower() == "forbidden":
                                print("Violating class: " + cls)
                            elif not hide:
                                print("Missing class: " + cls)
                            else:
                                print("Extra class: " + cls)
                else:
                    if len(cls.split("::")) != 1:
                            continue
                    if not classRestrict(newSource, 'at_least', cls, critData['scope'], True):
                        critAns = False
                        violationCount += 1
                        if not outputBool:
                            if not hide:
                                print("Missing class: " + cls)
                            else:
                                print("Extra class: " + cls)
                    exactCount += 1
                    if exactCount == 1:
                        if critData['scope'].lower() == "global" or critData['scope'] == "":
                            data = codeParser.prepareData(newSource)
                        else:
                            scopeG = scopeGetter(newSource, critData['scope'])
                            file = open("restrictorGen.cpp", "w")
                            file.write(scopeG)
                            file.close()
                            data = codeParser.prepareData("restrictorGen.cpp")
                        for decl in data['nodes']:
                            if decl['kind'] == "CLASS_DECL" and decl['start'] != decl['end']:
                                compareCount += 1

        #Handles functions in YAML file
        elif criteria == 'functions':
            for func in critData['names']:
                if critData['restriction'].lower() != 'exactly':
                    if not funcRestrict(newSource, critData['restriction'], func, critData['scope'], True):
                        critAns = False
                        violationCount += 1
                        if not outputBool:
                            if critData['restriction'].lower() == "forbidden":
                                print("Violating function: " + func)
                            elif not hide:
                                print("Missing function: " + func)
                            else:
                                print("Extra function: " + func)
                else:
                    if not funcRestrict(newSource, 'at_least', func, critData['scope'], True):
                        critAns = False
                        violationCount += 1
                        if not outputBool:
                            if not hide:
                                print("Missing function: " + func)
                            else:
                                print("Extra function: " + func)
                    exactCount += 1
                    if exactCount == 1:
                        if critData['scope'].lower() == "global" or critData['scope'] == "":
                             data = codeParser.prepareData(newSource)
                        else:
                            scopeG = scopeGetter(newSource, critData['scope'])
                            file = open("restrictorGen.cpp", "w")
                            file.write(scopeG)
                            file.close()
                            data = codeParser.prepareData("restrictorGen.cpp")
                        for decl in data['nodes']:
                            if decl['kind'] == "FUNCTION_DECL" and decl['start'] != decl['end']:
                                compareCount += 1

        #Handles (private/public/protected) functions in YAML file
        elif criteria == 'private_functions' or criteria == 'public_functions' or criteria == 'protected_functions':
            access = criteria.split("_")[0].upper()
            for func in critData['names']:
                if critData['restriction'].lower() != 'exactly':
                    if not accessRestrict(newSource, critData['restriction'], func, critData['scope'], access, True):
                        critAns = False
                        violationCount += 1
                        if not outputBool:
                            if critData['restriction'] == "forbidden":
                                print("Violating " + access + " function: " + func)
                            elif not hide:
                                print("Missing " + access + " function: " + func)
                            else:
                                print("Extra " + access + " function: " + func)
                else:
                    if not accessRestrict(newSource, 'at_least', func, critData['scope'], access, True):
                        critAns = False
                        violationCount += 1
                        if not outputBool:
                            if not hide:
                                print("Missing " + access + " function: " + func)
                            else:
                                print("Extra " + access + " function: " + func)
                    exactCount += 1
                    if exactCount == 1:
                        if critData['scope'].lower() == "global" or critData['scope'] == "":
                             data = codeParser.prepareData(newSource)
                        else:
                            scopeG = scopeGetter(newSource, critData['scope'])
                            file = open("restrictorGen.cpp", "w")
                            file.write(scopeG)
                            file.close()
                            data = codeParser.prepareData("restrictorGen.cpp")
                        for decl in data['nodes']:
                            if decl['kind'] == "CXX_METHOD" and decl['access_type'] == access and (decl['start'] != decl['end'] or decl['is_virtual_method'] == True):
                                if decl['is_virtual_method'] == True:
                                    if decl['parent_class'] in parentCheck:
                                        compareCount -= 1
                                    parentCheck.append(decl['parent_class'])
                                compareCount += 1
    

    #To print the last criteria in the loop
    if exactCount != 0 and compareCount != exactCount:
        critAns = critAns & False
    # if critData['restriction'].lower() == "exactly" and not critAns and not outputBool and not hide:
    #     print(critOld + " are not exactly the same.")
    if outputBool:
        print(violationCount)
    



#Function to restrict a single library, no need for the YAML file, further explanation available exactly below function definition
@app.command("l")
def libRestrict(source: str  = typer.Argument(..., help="The path of the .cpp or .h file the user wants restrictor to work on."),
                 restriction: str = typer.Argument(..., help="The restriction type used for 3 ways of checking:\n\nat_least: The library being checked must exist (It can be with other libraries).\n\nexactly: The library being checked must only exist (It can not be with other libraries).\n\nforbidden: The library being checked must not exist."),
                 lib: str = typer.Argument(..., help="The name of the library the user wants to check (only input the name of the library without #include)."),
                 hide:bool = typer.Argument(False, hidden=True, help="A hidden variable for developers use, used to return boolean")):
    """
    Checks if a certain library inputted by the user follows as certain restriction type of (at_least/exactly/forbidden) in a .cpp or .h file also inputted by the user.
    """

    #Makes sure that restriction inputted is a viable restriction
    if restriction.lower() not in ["exactly", "forbidden", "at_least"]:
        print("Invalid Restriction Input")
        return False
    
    with open(source, 'r') as f:
        source = f.read()

    #Handles restriction type "exactly"
    if restriction.lower() == "exactly":
        header = re.findall(r'^#include\s*[\s\S][<"]\S+[>"]$', source, flags=re.MULTILINE)
        if len(header) == 1:
            header = header[0].split("<")[1].split(">")[0] if len(header[0].split("<")) > 1 else header[0].split("\"")[1]
            if header == lib:
                if hide == False:
                    print("True")
                return True
            else:
                if hide == False:
                    print("False")
                return False

    #Handles restriction types "at_least" and "forbidden"
    else:
        existFlag = False
        header = re.findall(r'^#include\s*[\s\S][<"]\S+[>"]$', source, flags=re.MULTILINE)
        lib1 = "#include<" + lib + ">"
        lib2 = "#include\""+ lib + "\""

        for library in header:
            library = library.replace(" ","")
            if library == lib1 or library == lib2:
                existFlag = True

        if restriction.lower() == "forbidden":
            if existFlag == False:
                if not hide:
                    print("True")
                return True
            else:
                if not hide:
                    print("False")
                return False
        elif restriction.lower() == "at_least":
            if existFlag == True:
                if not hide:
                    print("True")
                return True
            else:
                if not hide:
                    print("False")
                return False



#Function to restrict a single keyword, no need for the YAML file, further explanation available exactly below function definition.
@app.command("k")
def wordRestrict(source: str  = typer.Argument(..., help="The path of the .cpp or .h file the user wants restrictor to work on."),
                 restriction: str = typer.Argument(..., help="The restriction type used for 2 ways of checking:\n\nat_least: The keyword being checked must exist (It can be with other keywords).\n\nforbidden: The keyword being checked must not exist.\n\nexactly not available for keywords, will work as at_least."),
                 keyword: str = typer.Argument(..., help="The keyword the user wants to check (This function matches using regex, function prototypes don't work here, must input exactly what you want to match)."),
                 scope: str = typer.Argument(..., help="The scope the user wants to check inside (Function or Class, use their prototypes)."),
                 hide:bool = typer.Argument(False, hidden=True, help="A hidden variable for developers use, used to return boolean")):
    """
    Checks if a certain keyword inputted by the user follows as certain restriction type of (at_least/exactly/forbidden) in a .cpp or .h file also inputted by the user.
    """

    #Makes sure that restriction inputted is a viable restriction
    if restriction.lower() not in ["exactly", "forbidden", "at_least"]:
        print("Invalid Restriction Input")
        return False

    #Check if global scope or not, if not then use scopeGetter to get everything in the scope defined by the user
    if scope.lower() == "global" or scope == "":
        with open(source, 'r') as f:
            source = f.read()
    else:
        source = scopeGetter(source, scope)
        if source == ["error"]:
            return False

    source = commentController.deleteForDeveloper(source)

    iter = 0
    if keyword.lower() == "iteration":
        iter += len(re.findall(fr"(?i){{[\s\S]*for[\s\S]*}}", source, flags=re.MULTILINE))
        iter += len(re.findall(fr"(?i){{[\s\S]*while[\s\S]*}}", source, flags=re.MULTILINE))
        
    #Check if keyword exists and print true or false according to restriction
    if len(re.findall(fr"(?i){{[\s\S]*{re.escape(keyword)}[\s\S]*}}", source, flags=re.MULTILINE)) > 0 or iter > 0:
        if restriction.lower() == "at_least" or restriction.lower() == "exactly":
            if not hide:
                print("True")
            return True
        elif restriction.lower() == "forbidden":
            if not hide:
                print("False")
            return False
    else:
        if restriction.lower() == "at_least" or restriction.lower() == "exactly":
            if not hide:
                print("False")
            return False
        elif restriction.lower() == "forbidden":
            if not hide:
                print("True")
            return True



#Function to restrict a single class, no need for the YAML file, further explanation available exactly below function definition.
@app.command("c")
def classRestrict(source: str  = typer.Argument(..., help="The path of the .cpp or .h file the user wants restrictor to work on."),
                 restriction: str = typer.Argument(..., help="The restriction type used for 3 ways of checking:\n\nat_least: The class being checked must exist (It can be with other classes).\n\nexactly: The class being checked must only exist (It can not be with other classes).\n\nforbidden: The class being checked must not exist."),
                 prototype: str = typer.Argument(..., help="The class the user wants to check (Must input like this: \"class name\")."),
                 scope: str = typer.Argument(..., help="The scope the user wants to check inside (Function or Class, input their prototypes)."),
                 hide:bool = typer.Argument(False, hidden=True, help="A hidden variable for developers use, used to return boolean")):
    """
    Checks if a certain class inputted by the user follows as certain restriction type of (at_least/exactly/forbidden) in a .cpp or .h file also inputted by the user.
    """

    #Makes sure that restriction inputted is a viable restriction
    if restriction.lower() not in ["exactly", "forbidden", "at_least"]:
        print("Invalid Restriction Input")
        return False
    
    #Check if global scope or not, if not then use scopeGetter to get everything in the scope defined by the user
    if scope.lower() == "global" or scope == "":
        data = codeParser.prepareData(source)
        with open(source, 'r') as f:
            scopeG = f.read()
    else:
        scopeG = scopeGetter(source, scope)
        if scopeG == ["error"]:
            return False
        file = open("restrictorGen.cpp", "w")
        file.write(scopeG)
        file.close()
        data = codeParser.prepareData("restrictorGen.cpp")

    prototype = prototype.strip()
    empty = []
    #Check if class exists and print true or false according to restriction
    if codeParser.findLocationClass(data, prototype, source, "class") != empty:
        if restriction.lower() == "exactly":
            if len(re.findall(r"(?:(?<=\s)|(?<=^))class.*?\{(?=\s|$)", scopeG, flags=re.MULTILINE)) > 1:
                if not hide:
                    print("False")
                return False
            else:
                if not hide:
                    print("True")
                return True
        elif restriction.lower() == "at_least":
            if not hide:
                print("True")
            return True
        elif restriction.lower() == "forbidden":
            if not hide:
                print("False")
            return False
    else:
        if restriction.lower() == "at_least" or restriction.lower() == "exactly":
            if not hide:
                print("False")
            return False
        elif restriction.lower() == "forbidden":
            if not hide:
                print("True")
            return True



#Function to restrict a single function, no need for the YAML file, further explanation available exactly below function definition.
@app.command("f")
def funcRestrict(source: str  = typer.Argument(..., help="The path of the .cpp or .h file the user wants restrictor to work on."),
                 restriction: str = typer.Argument(..., help="The restriction type used for 3 ways of checking:\n\nat_least: The function being checked must exist (It can be with other functions).\n\nexactly: The function being checked must only exist (It can not be with other functions).\n\nforbidden: The function being checked must not exist."),
                 prototype: str = typer.Argument(..., help="The function the user wants to check (Must input like this: \"int functionName(int, int)\"."),
                 scope: str = typer.Argument(..., help="The scope the user wants to check inside (Function or Class, input their prototypes, if class then no need to add class name to prototype)."),
                 hide:bool = typer.Argument(False, hidden=True, help="A hidden variable for developers use, used to return boolean")):
    """
    Checks if a certain function inputted by the user follows as certain restriction type of (at_least/exactly/forbidden) in a .cpp or .h file also inputted by the user.
    """
    
    #Makes sure that restriction inputted is a viable restriction
    if restriction.lower() not in ["exactly", "forbidden", "at_least"]:
        print("Invalid Restriction Input")
        return False
    
    #Check if global scope or not, if not then use scopeGetter to get everything in the scope defined by the user
    if scope.lower() == "global" or scope == "":
        data = codeParser.prepareData(source)
        with open(source, 'r') as f:
            scopeG = f.read()
    else:
        if len(scope.split("class ")) > 1:
            fname = prototype.split("(")[0].split(" ")[-1].strip()
            prototype = prototype.split(fname)[0] + scope.split("class ")[-1].strip() + "::" + fname + "(" + prototype.split("(")[-1]
        scopeG = scopeGetter(source, scope)
        if scopeG == ["error"]:
            return False
        file = open("restrictorGen.cpp", "w")
        file.write(scopeG)
        file.close()
        data = codeParser.prepareData("restrictorGen.cpp")

    empty = []
    #Check if function exists and print true or false according to restriction
    if codeParser.findLocationFunction(data, prototype, source) != empty:
        if restriction.lower() == "exactly":
            # if len(re.findall(r"\b[a-zA-Z_][a-zA-Z0-9_]*\s+[a-zA-Z_][a-zA-Z0-9_]*\s*\([^)]*\)\s*\{[^{}]*\}", scopeG, flags=re.MULTILINE)) > 1:
            if len(re.findall(r"\b(?:\w+\s+)+\*?\s*\w+\s*\([^)]*\)\s*(?:const\s*)?(?:=\s*0)?(?=\s*\{)", scopeG, flags=re.MULTILINE)) > 1:
                if not hide:
                    print("False")
                return False
            else:
                if not hide:
                    print("True")
                return True
        elif restriction.lower() == "at_least":
            if not hide:
                print("True")
            return True
        elif restriction.lower() == "forbidden":
            if not hide:
                print("False")
            return False
    else:
        if restriction.lower() == "at_least" or restriction.lower() == "exactly":
            if not hide:
                print("False")
            return False
        elif restriction.lower() == "forbidden":
            if not hide:
                print("True")
            return True



#Function to restrict a single (private/public/protected) function, no need for the YAML file, further explanation available exactly below function definition.
@app.command("a")
def accessRestrict(source: str  = typer.Argument(..., help="The path of the .cpp or .h file the user wants restrictor to work on."),
                 restriction: str = typer.Argument(..., help="The restriction type used for 3 ways of checking:\n\nat_least: The function being checked must exist (It can be with other functions).\n\nexactly: The function being checked must only exist (It can not be with other functions).\n\nforbidden: The function being checked must not exist."),
                 prototype: str = typer.Argument(..., help="The function the user wants to check (Must input like this: \"int functionName(int, int)\" or \"int functionName(int x,int y)\")."),
                 scope: str = typer.Argument(..., help="The scope the user wants to check inside (Function or Class, input their prototypes, if class then no need to add class name to prototype)."),
                 type: str = typer.Argument(..., help="The access type the user wants to check for (private/public/protected)."),
                 hide:bool = typer.Argument(False, hidden=True, help="A hidden variable for developers use, used to return boolean")):
    """
    Checks if a certain (private/public/protected) function inputted by the user follows as certain restriction type of (at_least/exactly/forbidden) in a .cpp or .h file also inputted by the user.
    """
    
    #Makes sure that restriction inputted is a viable restriction
    if restriction.lower() not in ["exactly", "forbidden", "at_least"]:
        print("Invalid Restriction Input")
        return False
    
    #Check if global scope or not, if not then use scopeGetter to get everything in the scope defined by the user as well as preparing data for access_type search
    if scope.lower() == "global" or scope == "":
        data = codeParser.prepareData(source)
    else:
        scopeG = scopeGetter(source, scope)
        if scopeG == ["error"]:
            return False
        file = open("restrictorGen.cpp", "w")
        file.write(scopeG)
        file.close()
        data = codeParser.prepareData("restrictorGen.cpp")

    unsigned = ""
    longlong = ""
    #Variables used in the following if statement to find if function is private
    if prototype.strip()[0:8] == "unsigned":
        unsigned = "unsigned "
    if len(prototype.split("long long")) > 1:
        longlong = "long "
    if len(prototype.split("virtual")) == 1 and len(prototype.split("static")) == 1:
        if unsigned == "":
            if longlong == "":
                proto = unsigned + longlong + prototype.split(' ')[0] + ' (' + prototype.split('(')[1].strip().split("=0")[0].strip()
            else:
                proto = unsigned + longlong + prototype.split(' ')[1] + ' (' + prototype.split('(')[1].strip().split("=0")[0].strip()
        else:
            if longlong == "":
                proto = unsigned + longlong + prototype.split(' ')[1] + ' (' + prototype.split('(')[1].strip().split("=0")[0].strip()
            else:
                proto = unsigned + longlong + prototype.split(' ')[2] + ' (' + prototype.split('(')[1].strip().split("=0")[0].strip()
        if len(proto.split("(")) > 2:
            proto = "void ()"
    else:
        if unsigned == "":
            if longlong == "":
                proto = unsigned + longlong + prototype.split(' ')[1] + ' (' + prototype.split('(')[1].strip().split("=0")[0].strip()
            else:
                proto = unsigned + longlong + prototype.split(' ')[2] + ' (' + prototype.split('(')[1].strip().split("=0")[0].strip()
        else:
            if longlong == "":
                proto = unsigned + longlong + prototype.split(' ')[2] + ' (' + prototype.split('(')[1].strip().split("=0")[0].strip()
            else:
                proto = unsigned + longlong + prototype.split(' ')[3] + ' (' + prototype.split('(')[1].strip().split("=0")[0].strip()

    unsigned = ""
    longlong = ""
    spelling = prototype.split('(')[0].split(' ')[-1].split('::')[-1]
    
    for decl in data['nodes']:
        if (decl['kind'] == "CXX_METHOD" or decl['kind'] == "CONSTRUCTOR" or decl['kind'] == "DESTRUCTOR") and decl['spelling'] == spelling and decl['prototype'].replace(" ","") == proto.replace(" ","") and decl['access_type'] == type.lower():
            return funcRestrict(source, restriction, prototype, scope, hide)
    if not hide:
        print("False")
    return False



#Function to restrict a single (private/public/protected) function, no need for the YAML file, further explanation available exactly below function definition.
def checkAPI(source: str  = typer.Argument(..., help="The path of the .cpp or .h file the user wants restrictor to work on."),
            restriction: str = typer.Argument(..., help="The restriction type used for 2 ways of checking:\n\nat_least: Everything being checked must exist (It can be with other functions/classes).\n\nexactly: Everything being checked must only exist (It can not be with other functions/classes)."),
            compare: str = typer.Argument(..., help="The path of the .cpp or .h file the user wants the source file to be compared to."),
            output: str  = typer.Option("n", "-o", help="If n this will make checkAPI print the number of missing functions/classes then extra functions/classes, Input V if you want a list of violations to be printed and more information (default is n) (Takes only v or V or n or N)."),
            hide:bool = typer.Argument(False, hidden=True, help="A hidden variable for developers use, used to return extra functions and classes found in the code.")):
    """
    Compares two files together, the source and compare file, it will check if the function prototypes and class names match then return true or false accordingly.
    """
    #Used to control the style of output
    if output not in ["n", "N", "V", "v"]:
        print("Invalid -o input")
        return False

    data = codeParser.prepareData(compare)

    #Makes sure that restriction inputted is a viable restriction
    if restriction.lower() not in ["exactly", "forbidden", "at_least"]:
        print("Invalid Restriction Input")
        return False
    
    #Varibales and loop used to create the YAML file needed to run restrictor, YAML file made from compare file.
    allPrivFunctions = []
    allPublicFunctions =[]
    allProtectedFunctions = []
    allFunctions = []
    allClasses = []
    virtual = ""
    const = ""
    unsigned = ""
    pure = ""
    static = ""
    for decl in data['nodes']:
        if decl['prototype'].split(' ')[0] == "unsigned":
            unsigned = " " + decl['prototype'].split(' ')[1]
        else:
            unsigned = ""
        if decl['kind'] == "CXX_METHOD":
            if decl['is_pure'] == True:
                pure = "=0"
            else:
                pure = ""
            if decl['is_static_method'] == True:
                static = "static "
            else:
                static = ""
            if decl['is_virtual_method'] == True:
                virtual = "virtual "
            else:
                virtual = ""
            if decl['is_const_method'] == True:
                const = " const"
            else:
                const = ""
            if decl['access_type'] == "":
                allFunctions.append(decl['prototype'].split(' ')[0] + unsigned + " " + decl['displayname'] + const + pure)
            elif decl['access_type'] == 'private':
                allPrivFunctions.append(virtual + static + decl['prototype'].split(' ')[0] + unsigned + " " + decl['parent_class'].split(' ')[1] + "::" + decl['displayname'] + const + pure)
            elif decl['access_type'] == 'public':
                allPublicFunctions.append(virtual + static + decl['prototype'].split(' ')[0] + unsigned + " " + decl['parent_class'].split(' ')[1] + "::" + decl['displayname'] + const + pure)
            elif decl['access_type'] == 'protected':
                allProtectedFunctions.append(virtual + static + decl['prototype'].split(' ')[0] + unsigned + " " + decl['parent_class'].split(' ')[1] + "::" + decl['displayname'] + const + pure)
        elif decl['kind'] == "FUNCTION_DECL":
            if len(decl['prototype'].split('(')[0].split('**')) == 1 and len(decl['prototype'].split('(')[0].split('*')) == 1:
                allFunctions.append(decl['prototype'].split(' ')[0] + unsigned + " " + decl['displayname'] + const + pure)
            elif len(decl['prototype'].split('**')) > 1:
                allFunctions.append(decl['prototype'].split('**')[0] + unsigned + "** " + decl['displayname'] + const + pure)
            else:
                allFunctions.append(decl['prototype'].split(' ')[0] + unsigned + " * " + decl['displayname' + const] + pure)
        if decl['kind'] == "CLASS_DECL":
            allClasses.append("class " + decl['prototype'])
    
    #Write the YAML file
    file_yaml = {'classes': {'restriction': f'{restriction}', 'scope': 'global', 'names': list(set(allClasses))}, 'functions': {'restriction': f'{restriction}', 'scope': 'global', 'names': list(set(allFunctions))}, 'public_functions': {'restriction': f'{restriction}', 'scope': 'global', 'names': list(set(allPublicFunctions))}, 'private_functions': {'restriction': f'{restriction}', 'scope': 'global', 'names': list(set(allPrivFunctions))}, 'protected_functions': {'restriction': f'{restriction}', 'scope': 'global', 'names': list(set(allProtectedFunctions))}}
    with open("checkAPI.yaml", 'w') as file:
        yaml.dump(file_yaml, file)
    
    restrict(source, "checkAPI.YAML", output, hide)

    if not hide:
        checkAPI(compare, restriction, source, output, True)



@app.callback()
def main():
    """
    Allows the user to restrict the use of certain criterion or many criteria in a source file.\n\nThe criteria are libraries, keywords, classes, functions and access type of functions.\n\n
    The restriction follows one of three types being at_least, exactly and forbidden.\n\n
    at_least: must exist, can be with others of the same criterion.\n\n
    exactly: must only exist, can not be with others of the same criterion.\n\n
    forbidden: must not exist.\n\n
    For more information about each tool simply add the command you need followed by --help, for example:\n\n restrict r --help
    """


if __name__ == "__main__":
    app()



import typer
from y4d import restrictor as restrictor

app = typer.Typer()

#Function to restrict a single (private/public/protected) function, no need for the YAML file, further explanation available exactly below function definition.
@app.command("")
def main(compare: str = typer.Argument(..., help="The path of the .cpp or .h file the user wants the source file to be compared to."),
         source: str  = typer.Argument(..., help="The path of the .cpp or .h file the user wants restrictor to work on."),
        restriction: str = typer.Argument(..., help="The restriction type used for 2 ways of checking:\n\nat_least: Everything being checked must exist (It can be with other functions/classes).\n\nexactly: Everything being checked must only exist (It can not be with other functions/classes)."),
        output: str  = typer.Option("n", "-o", help="If n this will make checkAPI print the number of missing functions/classes then extra functions/classes, Input V if you want a list of violations to be printed and more information (default is n) (Takes only v or V or n or N)."),
        hide:bool = typer.Argument(False, hidden=True, help="A hidden variable for developers use, used to return extra functions and classes found in the code.")):
    """
    Compares two files together, the source and compare file, it will check if the function prototypes and class names match then return true or false accordingly.
    """
    restrictor.checkAPI(source, restriction, compare, output, hide)


if __name__ == "__main__":
    app()